@@ -1,3 +1,4 @@
+import warnings
 from collections import OrderedDict
 from collections.abc import Mapping, Sequence
 from functools import partial
@@ -6,17 +7,17 @@ from urllib.parse import (SplitResult, parse_qsl,
                           urljoin, urlsplit, urlunsplit)
 
 from multidict import MultiDict, MultiDictProxy
+import idna
 
 
 from .quoting import quote, unquote
 
-__version__ = '0.10.0'
+__version__ = '0.18.0'
 
-__all__ = ['URL']
+__all__ = ('URL',)
 
 
 # is_leaf()
-# path normalization
 
 
 DEFAULT_PORTS = {
@@ -47,7 +48,7 @@ class cached_property:
         self.wrapped = wrapped
         try:
             self.__doc__ = wrapped.__doc__
-        except:  # pragma: no cover
+        except AttributeError:  # pragma: no cover
             self.__doc__ = ""
         self.name = wrapped.__name__
 
@@ -135,13 +136,14 @@ class URL:
     #               / path-noscheme
     #               / path-empty
     # absolute-URI  = scheme ":" hier-part [ "?" query ]
-    __slots__ = ('_cache', '_val', '_strict')
+    __slots__ = ('_cache', '_val')
 
-    def __init__(self, val='', *, encoded=False, strict=False):
+    def __init__(self, val='', *, encoded=False, strict=None):
+        if strict is not None:  # pragma: no cover
+            warnings.warn("strict parameter is ignored")
         if isinstance(val, URL):
             self._val = val._val
             self._cache = val._cache
-            self._strict = val._strict
             return
         if isinstance(val, str):
             val = urlsplit(val)
@@ -151,50 +153,34 @@ class URL:
         else:
             raise TypeError("Constructor parameter should be str")
 
-        self._strict = strict
-
         if not encoded:
             if not val[1]:  # netloc
                 netloc = ''
             else:
-                netloc = val.hostname
-                if netloc is None:
+                host = val.hostname
+                if host is None:
                     raise ValueError(
                         "Invalid URL: host is required for abolute urls.")
-                try:
-                    netloc.encode('ascii')
-                except UnicodeEncodeError:
-                    netloc = netloc.encode('idna').decode('ascii')
-                else:
-                    try:
-                        ip = ip_address(netloc)
-                    except:
-                        pass
-                    else:
-                        if ip.version == 6:
-                            netloc = '['+netloc+']'
-                if val.port:
-                    netloc += ':{}'.format(val.port)
-                if val.username:
-                    user = _quote(val.username)
-                    if val.password:
-                        user += ':' + _quote(val.password)
-                    netloc = user + '@' + netloc
+                netloc = self._make_netloc(val.username,
+                                           val.password,
+                                           host,
+                                           val.port,
+                                           encode=True)
+            path = _quote(val[2], safe='+@:', protected='/+')
+            if netloc:
+                path = self._normalize_path(path)
 
-            val = SplitResult(
-                val[0],  # scheme
-                netloc,
-                _quote(val[2], safe='+@:', protected='/+', strict=strict),
-                query=_quote(val[3], safe='=+&?/:@',
-                             protected=PROTECT_CHARS, qs=True, strict=strict),
-                fragment=_quote(val[4], safe='?/:@', strict=strict))
+            query = _quote(val[3], safe='=+&?/:@',
+                           protected=PROTECT_CHARS, qs=True)
+            fragment = _quote(val[4], safe='?/:@')
+            val = SplitResult(val[0], netloc, path, query, fragment)
 
         self._val = val
         self._cache = {}
 
     @classmethod
     def build(cls, *, scheme='', user='', password='', host='', port=None,
-              path='', query=None, query_string='', fragment='', strict=False):
+              path='', query=None, query_string='', fragment=''):
         """Creates and returns a new URL"""
 
         if host and not scheme:
@@ -207,15 +193,19 @@ class URL:
             raise ValueError(
                 "Only one of \"query\" or \"query_string\" should be passed")
 
+        netloc = cls._make_netloc(user, password, host, port, encode=True)
+        path = _quote(path, safe='@:', protected='/')
+        if netloc:
+            path = cls._normalize_path(path)
+
         url = cls(
             SplitResult(
                 scheme,
-                cls._make_netloc(user, password, host, port),
-                _quote(path, safe='@:', protected='/'),
+                netloc,
+                path,
                 _quote(query_string),
                 fragment
             ),
-            strict=strict,
             encoded=True
         )
 
@@ -236,12 +226,24 @@ class URL:
     def __eq__(self, other):
         if not isinstance(other, URL):
             return NotImplemented
-        return self._val == other._val
+
+        val1 = self._val
+        if not val1.path and self.is_absolute():
+            val1 = val1._replace(path='/')
+
+        val2 = other._val
+        if not val2.path and other.is_absolute():
+            val2 = val2._replace(path='/')
+
+        return val1 == val2
 
     def __hash__(self):
         ret = self._cache.get('hash')
         if ret is None:
-            ret = self._cache['hash'] = hash(self._val)
+            val = self._val
+            if not val.path and self.is_absolute():
+                val = val._replace(path='/')
+            ret = self._cache['hash'] = hash(val)
         return ret
 
     def __le__(self, other):
@@ -265,7 +267,7 @@ class URL:
         return self._val > other._val
 
     def __truediv__(self, name):
-        name = _quote(name, safe=':@', protected='/', strict=self._strict)
+        name = _quote(name, safe=':@', protected='/')
         if name.startswith('/'):
             raise ValueError("Appending path "
                              "starting from slash is forbidden")
@@ -278,19 +280,20 @@ class URL:
             parts = path.rstrip('/').split('/')
             parts.append(name)
             new_path = '/'.join(parts)
+        if self.is_absolute():
+            new_path = self._normalize_path(new_path)
         return URL(self._val._replace(path=new_path, query='', fragment=''),
                    encoded=True)
 
     def __getstate__(self):
-        return self._val, self._strict
+        return self._val,
 
     def __setstate__(self, state):
         if state[0] is None and isinstance(state[1], dict):
             # default style pickle
             self._val = state[1]['_val']
-            self._strict = state[1]['_strict']
         else:
-            self._val, self._strict = state
+            self._val, *unused = state
         self._cache = {}
 
     def is_absolute(self):
@@ -329,7 +332,8 @@ class URL:
         if not self._val.scheme:
             raise ValueError("URL should have scheme")
         v = self._val
-        netloc = self._make_netloc(None, None, v.hostname, v.port)
+        netloc = self._make_netloc(None, None, v.hostname, v.port,
+                                   encode=False)
         val = v._replace(netloc=netloc, path='', query='', fragment='')
         return URL(val, encoded=True)
 
@@ -361,7 +365,10 @@ class URL:
 
         """
         # not .username
-        return self._val.username
+        ret = self._val.username
+        if not ret:
+            return None
+        return ret
 
     @cached_property
     def user(self):
@@ -411,7 +418,10 @@ class URL:
         raw = self.raw_host
         if raw is None:
             return None
-        return raw.encode('ascii').decode('idna')
+        try:
+            return idna.decode(raw.encode('ascii'))
+        except UnicodeError:  # e.g. '::1'
+            return raw.encode('ascii').decode('idna')
 
     @property
     def port(self):
@@ -452,7 +462,8 @@ class URL:
         Empty value if URL has no query part.
 
         """
-        ret = MultiDict(parse_qsl(self.query_string, keep_blank_values=True))
+        ret = MultiDict(parse_qsl(self.raw_query_string,
+                                  keep_blank_values=True))
         return MultiDictProxy(ret)
 
     @property
@@ -475,12 +486,18 @@ class URL:
 
     @cached_property
     def path_qs(self):
-        """Decoded path of URL with query
-        """
+        """Decoded path of URL with query."""
         if not self.query_string:
             return self.path
         return '{}?{}'.format(self.path, self.query_string)
 
+    @cached_property
+    def raw_path_qs(self):
+        """Encoded path of URL with query."""
+        if not self.raw_query_string:
+            return self.raw_path
+        return '{}?{}'.format(self.raw_path, self.raw_query_string)
+
     @property
     def raw_fragment(self):
         """Encoded fragment part of URL.
@@ -564,14 +581,68 @@ class URL:
         return _unquote(self.raw_name)
 
     @classmethod
-    def _make_netloc(cls, user, password, host, port):
-        ret = host
+    def _normalize_path(cls, path):
+        # Drop '.' and '..' from path
+
+        segments = path.split('/')
+        resolved_path = []
+
+        for seg in segments:
+            if seg == '..':
+                try:
+                    resolved_path.pop()
+                except IndexError:
+                    # ignore any .. segments that would otherwise cause an
+                    # IndexError when popped from resolved_path if
+                    # resolving for rfc3986
+                    pass
+            elif seg == '.':
+                continue
+            else:
+                resolved_path.append(seg)
+
+        if segments[-1] in ('.', '..'):
+            # do some post-processing here.
+            # if the last segment was a relative dir,
+            # then we need to append the trailing '/'
+            resolved_path.append('')
+
+        return '/'.join(resolved_path)
+
+    @classmethod
+    def _encode_host(cls, host):
+        try:
+            host = idna.encode(host, uts46=True).decode('ascii')
+        except UnicodeError:
+            host = host.encode('idna').decode('ascii')
+        try:
+            ip = ip_address(host)
+        except ValueError:
+            pass
+        else:
+            if ip.version == 6:
+                host = '[' + host + ']'
+        return host
+
+    @classmethod
+    def _make_netloc(cls, user, password, host, port, encode):
+        if encode:
+            ret = cls._encode_host(host)
+        else:
+            ret = host
         if port:
             ret = ret + ':' + str(port)
         if password:
             if not user:
-                raise ValueError("Non-empty password requires non-empty user")
+                user = ''
+            else:
+                if encode:
+                    user = _quote(user)
+            if encode:
+                password = _quote(password)
             user = user + ':' + password
+        elif user and encode:
+            user = _quote(user)
         if user:
             ret = user + '@' + ret
         return ret
@@ -609,7 +680,8 @@ class URL:
         return URL(self._val._replace(netloc=self._make_netloc(user,
                                                                password,
                                                                val.hostname,
-                                                               val.port)),
+                                                               val.port,
+                                                               encode=False)),
                    encoded=True)
 
     def with_password(self, password):
@@ -636,7 +708,8 @@ class URL:
                 netloc=self._make_netloc(val.username,
                                          password,
                                          val.hostname,
-                                         val.port)),
+                                         val.port,
+                                         encode=False)),
             encoded=True)
 
     def with_host(self, host):
@@ -656,19 +729,14 @@ class URL:
                              "for relative URLs")
         if not host:
             raise ValueError("host removing is not allowed")
-        try:
-            ip = ip_address(host)
-        except:
-            host = host.encode('idna').decode('ascii')
-        else:
-            if ip.version == 6:
-                host = '['+host+']'
+        host = self._encode_host(host)
         val = self._val
         return URL(
             self._val._replace(netloc=self._make_netloc(val.username,
                                                         val.password,
                                                         host,
-                                                        val.port)),
+                                                        val.port,
+                                                        encode=False)),
             encoded=True)
 
     def with_port(self, port):
@@ -689,28 +757,22 @@ class URL:
             self._val._replace(netloc=self._make_netloc(val.username,
                                                         val.password,
                                                         val.hostname,
-                                                        port)),
+                                                        port,
+                                                        encode=False)),
             encoded=True)
 
-    def with_path(self, path, encoded=False):
+    def with_path(self, path, *, encoded=False):
         """Return a new URL with path replaced."""
         if not encoded:
-            path = _quote(path, safe='@:', protected='/', strict=self._strict)
-        return URL(self._val._replace(path=path), encoded=True)
-
-    def with_query(self, *args, **kwargs):
-        """Return a new URL with query part replaced.
-
-        Accepts any Mapping (e.g. dict, multidict.MultiDict instances)
-        or str, autoencode the argument if needed.
-
-        It also can take an arbitrary number of keyword arguments.
-
-        Clear query if None is passed.
-
-        """
-        # N.B. doesn't cleanup query/fragment
+            path = _quote(path, safe='@:', protected='/')
+            if self.is_absolute():
+                path = self._normalize_path(path)
+        if len(path) > 0 and path[0] != '/':
+            path = '/' + path
+        return URL(self._val._replace(path=path, query='', fragment=''),
+                   encoded=True)
 
+    def _get_str_query(self, *args, **kwargs):
         if kwargs:
             if len(args) > 0:
                 raise ValueError("Either kwargs or single query parameter "
@@ -725,7 +787,7 @@ class URL:
         if query is None:
             query = ''
         elif isinstance(query, Mapping):
-            quoter = partial(_quote, qs=True, strict=self._strict)
+            quoter = partial(_quote, qs=True)
             lst = []
             for k, v in query.items():
                 if isinstance(v, str):
@@ -735,56 +797,59 @@ class URL:
                 else:
                     raise TypeError("Invalid variable type: mapping value "
                                     "should be str or int, got {!r}".format(v))
-                lst.append(quoter(k, safe='/?:@')+'='+quoter(v, safe='/?:@;'))
-                query = '&'.join(lst)
+                lst.append(
+                    quoter(k, safe='/?:@') + '=' + quoter(v, safe='/?:@'))
+            query = '&'.join(lst)
         elif isinstance(query, str):
             query = _quote(query, safe='/?:@',
                            protected=PROTECT_CHARS,
-                           qs=True, strict=self._strict)
+                           qs=True)
         elif isinstance(query, (bytes, bytearray, memoryview)):
             raise TypeError("Invalid query type: bytes, bytearray and "
                             "memoryview are forbidden")
         elif isinstance(query, Sequence):
-            quoter = partial(_quote, qs=True, strict=self._strict)
-            query = '&'.join(quoter(k, safe='/?:@')+'='+quoter(v, safe='/?:@;')
+            quoter = partial(_quote, qs=True, safe='/?:@')
+            query = '&'.join(quoter(k) + '=' + quoter(v)
                              for k, v in query)
         else:
             raise TypeError("Invalid query type: only str, mapping or "
                             "sequence of (str, str) pairs is allowed")
-        path = self._val.path
-        if path == '':
-            path = '/'
+
+        return query
+
+    def with_query(self, *args, **kwargs):
+        """Return a new URL with query part replaced.
+
+        Accepts any Mapping (e.g. dict, multidict.MultiDict instances)
+        or str, autoencode the argument if needed.
+
+        It also can take an arbitrary number of keyword arguments.
+
+        Clear query if None is passed.
+
+        """
+        # N.B. doesn't cleanup query/fragment
+
+        new_query = self._get_str_query(*args, **kwargs)
         return URL(
-            self._val._replace(path=path, query=query), encoded=True)
+            self._val._replace(path=self._val.path, query=new_query),
+            encoded=True)
 
     def update_query(self, *args, **kwargs):
         """Return a new URL with query part updated."""
-        if kwargs:
-            if len(args) > 0:
-                raise ValueError("Either kwargs or single query parameter "
-                                 "must be present")
-            new_query = kwargs
-        elif len(args) == 1:
-            new_query = args[0]
-
-            if isinstance(new_query, str):
-                new_query = OrderedDict(
-                    map(
-                        lambda x: x.split('=', 1),
-                        _quote(new_query,
-                               safe='/?:@', protected=PROTECT_CHARS,
-                               qs=True,
-                               strict=self._strict).lstrip("?").split("&")
-                        )
-                )
-
-        else:
-            raise ValueError("Either kwargs or single query parameter "
-                             "must be present")
-
+        new_query = OrderedDict(
+            map(
+                lambda x: x.split('=', 1),
+                _quote(self._get_str_query(*args, **kwargs),
+                       safe='/?:@', protected=PROTECT_CHARS,
+                       qs=True).lstrip("?").split("&")
+            )
+        )
         query = OrderedDict(self.query)
         query.update(new_query)
-        return self.with_query(tuple(query.items()))
+        query = self._get_str_query(query)
+        return URL(
+            self._val._replace(path=self._val.path, query=query), encoded=True)
 
     def with_fragment(self, fragment):
         """Return a new URL with fragment replaced.
@@ -801,7 +866,7 @@ class URL:
             raise TypeError("Invalid fragment type")
         return URL(
             self._val._replace(
-                fragment=_quote(fragment, safe='?/:@', strict=self._strict)),
+                fragment=_quote(fragment, safe='?/:@')),
             encoded=True)
 
     def with_name(self, name):
@@ -818,6 +883,8 @@ class URL:
         if '/' in name:
             raise ValueError("Slash in name is not allowed")
         name = _quote(name, safe='@:', protected='/')
+        if name in ('.', '..'):
+            raise ValueError(". and .. values are forbidden")
         parts = list(self.raw_parts)
         if self.is_absolute():
             if len(parts) == 1:
@@ -857,7 +924,8 @@ class URL:
                                       self._make_netloc(self.user,
                                                         self.password,
                                                         self.host,
-                                                        self._val.port),
+                                                        self._val.port,
+                                                        encode=False),
                                       self.path,
                                       self.query_string,
                                       self.fragment))
